module App exposing (..)

import Array exposing (Array)
import Browser
import Browser.Events
import Grid exposing (Coords)
import Html as H exposing (Html)
import Html.Attributes as HA
import Html.Events as HE
import Json.Decode as JD
import Random
import Tuple exposing (pair, first, second)

dynamite_cost level = (level+1) * 5

money_for level = 
  Array.get level
    (Array.fromList [0,5,10,20,50,75,100])
  |> Maybe.withDefault (level * 50 - 200)
{- 4, 9,  16, 25 -}
{- 5, 10, 20, 50, 100  -}

main = Browser.document
    { init = init
    , update = update
    , subscriptions = subscriptions
    , view = view
    }

type Action
  = Dig
  | Fill

type Rock
  = Dirt
  | Boulder
  | Treasure

type alias Cell =
  { dug : Int
  , rocks : Array (Rock, Bool)
  }

cell_rock : Cell -> Rock
cell_rock cell = (Array.get cell.dug cell.rocks) |> Maybe.map first |> Maybe.withDefault Dirt

type alias Grid = Grid.Grid Cell

type alias Model = 
  { grid : Grid
  , money : Int
  , action : Action
  }

default_cell = 
  { dug = 0
  , rocks = (Array.repeat 10 (Dirt, False))
  }

empty_grid : Grid
empty_grid = Grid.fill 5 5 default_cell

rock_probabilities =
  [ (10, Boulder)
  ]

random_rock : Random.Generator (Rock, Bool) 
random_rock =
  Random.weighted
    (100, Dirt)
    rock_probabilities

  |> Random.map (\r -> (r, False))

random_walk : Random.Generator (List Coords)
random_walk =
  let
    start = Random.map2 pair (Random.int 0 9) (Random.int 0 9)

    step : Int -> (Random.Generator (List Coords)) -> (Random.Generator (List Coords))
    step _ rlast = 
      rlast |> Random.andThen 
        (\last -> case last of
          [] -> Random.map List.singleton start
          (x,y)::others -> 
            let
              choices = [(x-1,y),(x+1,y),(x,y-1),(x,y+1)] |> List.filter (\(x2,y2) -> x>=0 && x<10 && y>=0 && y<10)
            in
              Random.uniform (x,y) choices |> Random.map (\c -> c::last)
        )
  in
    List.foldl
      step
      (Random.constant [])
      (List.range 0 10)

random_grid : Random.Generator Grid
random_grid =
  let
    width = 7
    height = 7
    
    rcells = 
      Random.list 
        (width*height)
        (    Random.list 10 random_rock
          |> Random.map (Array.fromList >> (\rocks -> { default_cell | rocks = rocks }))
        )
      |> Random.map Array.fromList
      
    rgrid =
         rcells
      |> Random.map (\cells -> { width = width, height = height, cells = cells })

    rwalks : Random.Generator (List (List Coords))
    rwalks = Random.list 10 random_walk

    apply_walk : List Coords -> Int -> Grid -> Grid
    apply_walk walk start grid =
      List.foldl
        (\(z,pos) -> Grid.map_cell pos (\cell -> { cell | rocks = Array.set (z+start) (Treasure, False) cell.rocks}))
        grid
        (List.indexedMap pair walk)
  in
    Random.andThen
      (\walks ->
        List.foldl
          (\walk -> Random.map2 (apply_walk walk) (Random.int 1 3))
          rgrid
          walks
      )
      rwalks

init_model : Model
init_model = 
  { grid = empty_grid 
  , money = 0
  , action = Dig
  }

type Msg
  = SelectCell Int Coords
  | SetGrid Grid
  | SetAction Action
  | Noop

nocmd model = (model, Cmd.none)

init : () -> (Model, Cmd Msg)
init _ = (init_model, Random.generate SetGrid random_grid)

update msg model = case (Debug.log "msg" msg) of
  SelectCell button coords -> click_cell button coords model |> nocmd
  
  SetGrid grid -> { model | grid = Debug.log "grid" grid } |> nocmd

  SetAction action -> { model | action = action } |> nocmd

  Noop -> model |> nocmd

click_result : Coords -> Model -> Maybe (Cell, Int)
click_result coords model =
  let
    grid = model.grid
    cell = Grid.get_cell coords grid |> Maybe.withDefault default_cell
    neighbours = Grid.neighbours 1 coords model.grid
    num_equal = neighbours |> List.filter (.dug >> (==) cell.dug) |> List.length
    sum_diff = neighbours |> List.map (\c2 -> c2.dug - cell.dug) |> List.sum
    all_greater = neighbours |> List.all (.dug >> (<) cell.dug)
    rock = cell_rock cell
    res : Maybe (Cell, Int)
    res =
      case (rock, model.action) of
        (Treasure, Dig) -> Just ({ cell | rocks = Array.set cell.dug (Dirt,True) cell.rocks }, money_for cell.dug)
        
        (Dirt, Dig) -> 
          if num_equal >= cell.dug + 1 || sum_diff > 2 || all_greater || cell.dug == 0 then
            Just ({ cell | dug = cell.dug + 1 }, cell.dug)
          else
            Nothing
            
        (Boulder, Dig) -> 
          Just ({ cell | rocks = Array.set cell.dug (Dirt, False) cell.rocks }, -(dynamite_cost cell.dug))
            
        (_, Fill) -> 
          if cell.dug > 0 then
            Just ({ cell | dug = cell.dug - 1 }, -(cell.dug-1))
          else
            Nothing
  in
    res |> Maybe.andThen (\(ncell, cost) -> if model.money + cost >= 0 then res else Nothing)

button_action button action = case button of
  0 -> action
  _ -> case action of
    Dig -> Fill
    _ -> Dig

click_cell button coords model =
  case click_result coords { model | action = button_action button model.action } of
    Nothing -> model
    Just (ncell, money) ->
      let
        ngrid = Grid.set_cell coords ncell model.grid
      in
        { model | grid = ngrid, money = model.money + money }

can_action_cell : Coords -> Model -> Bool
can_action_cell coords model = click_result coords model /= Nothing

total_score : Grid -> Int
total_score =
    (.cells)
  >> Array.toList
  >> List.map (.dug)
  >> List.sum

subscriptions model = 
  Browser.Events.onKeyDown
    (   JD.field "key" JD.string
     |> JD.andThen (\key ->
          case (String.toUpper key) of
            "D" -> JD.succeed (SetAction Dig)
            "F" -> JD.succeed (SetAction Fill)
            _ -> JD.fail (Debug.log "bad" ("unrecognised key "++key)) 
        )
    )


radio : List (H.Attribute Msg) -> String -> Html Msg
radio attr label =
  H.label
    []
    [ H.input
        (([ HA.type_ "radio"])++attr)
        []
    , H.text label
    ]

view : Model -> Browser.Document Msg
view model = 
  let
    grid = model.grid

    view_cell : Coords -> Cell -> Html Msg
    view_cell coords v =
      let
        (rock, flag) = Array.get v.dug v.rocks |> Maybe.withDefault (Dirt, False)
      in
        H.td 
          [ HA.classList
             [ ("can-increment", can_action_cell coords model)
             ]
          , HA.style "background" <| "hsl(60,50%,"++(String.fromFloat (50 / (toFloat (v.dug+1))))++"%)" 
          , HA.attribute "data-background" <| "hsl(60,50%,"++(String.fromFloat (50 / (toFloat (v.dug+1))))++")" 
          , HE.on "mousedown" 
              (    JD.field "button" JD.int 
                |> JD.map (\button ->
                  let
                    z = Debug.log "button" button
                  in
                    SelectCell button coords
                  )
              )
          , HE.preventDefaultOn "contextmenu" <| JD.succeed (Noop, True)
          ]
          [ H.div
            []
            [ H.text <| 
              (
              case rock of
                Dirt -> " "
                Boulder -> "🪨"
                Treasure -> "🪙"
              ) ++ (
                String.fromInt v.dug
              ) ++ (
                if flag then "⚐" else " "
              )
            ]
          ]

    view_grid : Html Msg
    view_grid =
      H.table [] 
        (   List.range 0 (grid.height-1) 
        |>  List.map  
              (\y -> 
                H.tr
                  [] 
                  (List.range 0 (grid.width-1) 
                  |> List.map 
                      (\x -> 
                        view_cell (x,y) (Grid.get_cell (x,y) grid |> Maybe.withDefault default_cell)
                      )
                  )
              )
        )
    
    action_radio action label =
      radio
        [ HE.onClick <| SetAction action
        , HA.value label
        , HA.name "action"
        , HA.checked <| model.action == action
        ]
        label
  in
    {
      title = "Hey",
      body = 
        [ view_grid 
                    
        , H.p
            []
            [ H.text <| ("Score: "++(String.fromInt <| total_score model.grid))
            ]
        , H.p
            []
            [ H.text <| ("Money: £"++(String.fromInt <| model.money))
            ] 
        , H.fieldset
          [ ]
          [ H.legend
            []
            [ H.text <| "Action: " ++ (Debug.toString model.action) ]
          , action_radio Dig "Dig / Fill"
          , action_radio Fill "Fill / Dig"
          ]
        ]
    }